using InlineExports

# Enum defining the two implemented boundary conditions.
@enum BC_TYPE begin
    BC_TYPE_CLOSED
    BC_TYPE_CONSTANT
end
@exportinstances BC_TYPE

# Enum defining all 4 possible sides to a 1D and 2D grid.
@enum BC_SIDE begin
    BC_SIDE_LEFT
    BC_SIDE_RIGHT
    BC_SIDE_TOP
    BC_SIDE_BOTTOM
end
@exportinstances BC_SIDE

# This class defines the boundary conditions of individual boundary elements.
# These can be flexibly used and combined later in other classes.
# The class serves as an auxiliary class for structuring the Boundary class.
#
# T is the data type of the boundary condition element
@export mutable struct BoundaryElement{T<:Number}
    type::BC_TYPE
    value::T

    # Construct a new Boundary Element object for the closed case. The boundary type is here
    # automatically set to the type BC_TYPE_CLOSED, where the value takes -1 and does not hold
    # any physical meaning.
    function BoundaryElement{T}() where {T<:Number}
        new(BC_TYPE_CLOSED, -1)
    end

    # Construct a new Boundary Element object for the constant case. The boundary type is
    # automatically set to the type BC_TYPE_CONSTANT.
    #
    # Arguments:
    #   value: Value of the constant concentration to be assumed at the corresponding
    #          boundary element.
    function BoundaryElement{T}(value::T) where {T<:Number}
        new(BC_TYPE_CONSTANT, value)
    end
end

# Allows changing the boundary type of a corresponding BoundaryElement object.
#
# Arguments:
#   type: Type of boundary condition. Either BC_TYPE_CONSTANT or BC_TYPE_CLOSED.
@export function setBoundaryElementType!(this::BoundaryElement{T}, type::BC_TYPE) where {T}
    this.type = type
end

# Sets the value of a boundary condition for the constant case.
#
# Arguments:
#   value: Concentration to be considered constant for the corresponding boundary element.
@export function setBoundaryElementValue!(this::BoundaryElement{T}, value::T) where {T}
    if this.type == BC_TYPE_CLOSED
        error(
            "No constant boundary concentrations can be set for closed boundaries. Please change type first.",
        )
    end

    this.value = value
end

# Return the type of the boundary condition, i.e. whether the boundary is considered closed or constant.
#
# Returns: Type of boundary condition, either BC_TYPE_CLOSED or BC_TYPE_CONSTANT.
@export function getBoundaryElementType(this::BoundaryElement{T})::BC_TYPE where {T}
    return this.type
end

# Return the concentration value for the constant boundary condition.
#
# Returns: Value of the concentration.
@export function getBoundaryElementValue(this::BoundaryElement{T})::T where {T}
    return this.value
end

@export mutable struct Boundary{T<:Number}
    dim::UInt8
    cols::UInt32
    rows::UInt32

    # Vector with Boundary Element information
    boundaries::Dict{BC_SIDE,Vector{BoundaryElement{T}}}

    # Inner boundary conditions for 1D and 2D grids identified by (row,col)
    inner_boundary::Dict{Tuple{UInt32,UInt32},T}

    # Creates a boundary object for a 1D grid
    #
    # Arguments:
    #   length: Length of the grid
    function Boundary{T}(length::UInt32) where {T}
        Boundary{T}(Grid{T}(Int(length)))
    end

    # Creates a boundary object for a 2D grid
    #
    # Arguments:
    #   rows: Number of rows of the grid
    #   cols: Number of columns of the grid
    function Boundary{T}(rows::UInt32, cols::UInt32) where {T}
        Boundary{T}(Grid{T}(Int(rows), Int(cols)))
    end

    # Creates a boundary object based on the passed grid object and initializes the boundaries as closed.
    #
    # Arguments:
    #   grid: Grid object on the basis of which the simulation takes place and from which the
    #         dimensions (in 2D case) are taken.
    function Boundary{T}(grid::Grid{T}) where {T}
        dim::UInt8 = getGridDim(grid)
        cols::UInt32 = getGridCol(grid)
        rows::UInt32 = getGridRow(grid)

        boundaries = Dict{BC_SIDE,Vector{BoundaryElement{T}}}()
        inner_boundary = Dict{Tuple{UInt32,UInt32},T}()

        if dim == 1
            # in 1D only left and right boundary
            boundaries[BC_SIDE_LEFT] = [BoundaryElement{T}()]
            boundaries[BC_SIDE_RIGHT] = [BoundaryElement{T}()]
        elseif dim == 2
            boundaries[BC_SIDE_LEFT] = [BoundaryElement{T}() for _ = 1:rows]
            boundaries[BC_SIDE_RIGHT] = [BoundaryElement{T}() for _ = 1:rows]
            boundaries[BC_SIDE_TOP] = [BoundaryElement{T}() for _ = 1:cols]
            boundaries[BC_SIDE_BOTTOM] = [BoundaryElement{T}() for _ = 1:cols]
        end

        new(dim, cols, rows, boundaries, inner_boundary)
    end
end

# Sets all elements of the specified boundary side to the boundary condition closed.
#
# Arguments:
#   side: Side to be set to closed, e.g. BC_SIDE_LEFT.
@export function setBoundarySideClosed!(this::Boundary{T}, side::BC_SIDE) where {T}
    if this.dim == 1
        if side == BC_SIDE_BOTTOM || side == BC_SIDE_TOP
            error(
                "For the one-dimensional case, only the BC_SIDE_LEFT and BC_SIDE_RIGHT borders exist.",
            )
        end
    end

    is_vertical::Bool = side == BC_SIDE_LEFT || side == BC_SIDE_RIGHT
    n::Int = is_vertical ? this.rows : this.cols

    this.boundaries[side] = [BoundaryElement{T}() for _ = 1:n]
end

# Sets all elements of the specified boundary side to the boundary condition constant.
# Thereby the concentration values of the boundaries are set to the passed value.
#
# Arguments:
#   side: Side to be set to constant, e.g. BC_SIDE_LEFT.
#   value: Concentration to be set for all elements of the specified page.
@export function setBoundarySideConstant!(this::Boundary{T}, side::BC_SIDE, value::T) where {T}
    if this.dim == 1
        if side == BC_SIDE_BOTTOM || side == BC_SIDE_TOP
            error(
                "For the one-dimensional case, only the BC_SIDE_LEFT and BC_SIDE_RIGHT borders exist.",
            )
        end
    end

    is_vertical::Bool = side == BC_SIDE_LEFT || side == BC_SIDE_RIGHT
    n::Int = is_vertical ? this.rows : this.cols

    this.boundaries[side] = [BoundaryElement{T}(value) for _ = 1:n]
end

# Specifically sets the boundary element of the specified side defined by the index to the
# boundary condition closed.
#
# Arguments:
#   side: Side in which an element is to be defined as closed.
#   index: Index of the boundary element on the corresponding boundary side.
#          Must index an element of the corresponding side.
@export function setBoundaryElementClosed!(this::Boundary{T}, side::BC_SIDE, index::Int) where {T}
    # tests whether the index really points to an element of the boundary side.
    if (size(this.boundaries[side], 1) < index) || index < 1
        error("Index is selected either too large or too small.")
    end

    setBoundaryElementType!(this.boundaries[side][index], BC_TYPE_CLOSED)
end

# Specifically sets the boundary element of the specified side defined by the index to the boundary
# condition constant with the given concentration value.
#
# Arguments:
#   side: Side in which an element is to be defined as constant.
#   index: Index of the boundary element on the corresponding boundary side.
#          Must index an element of the corresponding side.
#   value: Concentration value to which the boundary element should be set.
@export function setBoundaryElementConstant!(
    this::Boundary{T},
    side::BC_SIDE,
    index::Int,
    value::T,
) where {T}
    # tests whether the index really points to an element of the boundary side.
    if (size(this.boundaries[side], 1) < index) || index < 1
        error("Index is selected either too large or too small.")
    end

    setBoundaryElementType!(this.boundaries[side][index], BC_TYPE_CONSTANT)
    setBoundaryElementValue!(this.boundaries[side][index], value)
end

# Returns the boundary condition of a specified side as a vector of BoundarsElement objects.
#
# Arguments:
#   side: Boundary side from which the boundary conditions are to be returned.
# Returns: Contains the boundary conditions as BoundaryElement<T> objects.
@export function getBoundarySide(
    this::Boundary{T},
    side::BC_SIDE,
)::AbstractVector{BoundaryElement{T}} where {T}
    if this.dim == 1
        if side == BC_SIDE_BOTTOM || side == BC_SIDE_TOP
            error(
                "For the one-dimensional trap, only the BC_SIDE_LEFT and BC_SIDE_RIGHT borders exist.",
            )
        end
    end

    return this.boundaries[side]
end

# Returns the boundary condition of a specified element on a given side.
#
# Arguments:
#   side: Boundary side in which the boundary condition is located.
#   index: Index of the boundary element on the corresponding boundary side.
#          Must index an element of the corresponding side.
# Return: Boundary condition as a BoundaryElement{T} object.
@export function getBoundaryElement(
    this::Boundary{T},
    side::BC_SIDE,
    index::Int,
)::BoundaryElement{T} where {T}
    # tests whether the index really points to an element of the boundary side.
    if (size(this.boundaries[side], 1) < index) || index < 1
        error("Index is selected either too large or too small.")
    end

    return this.boundaries[side][index]
end

# Returns the type of a boundary condition, i.e. either BC_TYPE_CLOSED or BC_TYPE_CONSTANT.
#
# Arguments:
#   side: Boundary side in which the boundary condition is located.
#   index: Index of the boundary element on the corresponding boundary side.
#          Must index an element of the corresponding side.
# Return: Boundary Type of the corresponding boundary condition.
@export function getBoundaryElementType(
    this::Boundary{T},
    side::BC_SIDE,
    index::Int,
)::BC_TYPE where {T}
    # tests whether the index really points to an element of the boundary side.
    if (size(this.boundaries[side], 1) < index) || index < 1
        error("Index is selected either too large or too small.")
    end

    return getBoundaryElementType(this.boundaries[side][index])
end

# Returns the concentration value of a corresponding BoundaryElement{T}
# object if it is a constant boundary condition.
#
# Arguments:
#   side: Boundary side in which the boundary condition is located.
#   index: Index of the boundary element on the corresponding boundary side.
#          Must index an element of the corresponding side.
# Return: Concentration of the corresponding BoundaryElement{T} object.
@export function getBoundaryElementValue(this::Boundary{T}, side::BC_SIDE, index::Int)::T where {T}
    # tests whether the index really points to an element of the boundary side.
    if (size(this.boundaries[side], 1) < index) || index < 1
        error("Index is selected either too large or too small.")
    end

    return getBoundaryElementValue(this.boundaries[side][index])
end

# Get thes Boundary Side Values as a vector. Value is -1 in case some specific boundary is closed.
#
# Arguments:
#   side: Boundary side for which the values are to be returned.
# Returns: Vector with values as T.
@export function getBoundarySideValues(
    this::Boundary{T},
    side::BC_SIDE,
)::AbstractVector{T} where {T}
    length::Int = size(this.boundaries[side], 1)
    values = Vector{T}(undef, length)

    for i = 1:length
        if getBoundaryElementType(this, side, i) == BC_TYPE_CLOSED
            values[i] = -1
            continue
        end

        values[i] = getBoundaryElementValue(this, side, i)
    end

    return values
end

# Serializes the boundary conditions into a vector of BoundaryElement objects.
#
# Returns: Vector with BoundaryElement objects.
@export function serializeBoundary(this::Boundary{T})::AbstractVector{BoundaryElement{T}} where {T}
    serialized::Vector{BoundaryElement{T}} = []

    for side in keys(this.boundaries)
        for index in eachindex(this.boundaries[side])
            append!(serialized, this.boundaries[side][index])
        end
    end

    return serialized
end

# Deserializes the boundary conditions from a vector of BoundaryElement objects.
#
# Arguments:
#   serialized: Vector with BoundaryElement objects
@export function deserializeBoundary!(
    this::Boundary{T},
    serialized::AbstractVector{BoundaryElement{T}},
) where {T}
    index::Int = 0

    for side in keys(this.boundaries)
        for i in eachindex(this.boundaries[side])
            this.boundaries[side][i] = serialized[index]
            index += 1
        end
    end
end

# Arguments:
#   index: Index of the inner constant boundary condition
#   value: Value of the inner constant boundary condition
@export function setInnerBoundary!(this::Boundary{T}, index::Int, value::T) where {T}
    if this.dim != 1
        error("This function is only available for 1D grids.")
    end

    if index > this.cols
        error("Index is out of bounds.")
    end

    this.inner_boundary[(1, index)] = value
end

# Set inner constant boundary condition in 2D case.
#
# Arguments:
#   row: Row index of the inner constant boundary condition
#   col: Column index of the inner constant boundary condition
#   value: Value of the inner constant boundary condition
@export function setInnerBoundary!(this::Boundary{T}, row::Int, col::Int, value::T) where {T}
    if this.dim != 2
        error("This function is only available for 2D grids.")
    end

    if row > this.rows || col > this.cols
        error("Index is out of bounds.")
    end

    this.inner_boundary[(row, col)] = value
end

# Get inner constant boundary condition in 1D case.
#
# Arguments:
#   index: Index of the inner constant boundary condition
# Returns: Pair of boolean (whether constant boundary was set or not) and
#          value of the inner constant boundary condition.
@export function getInnerBoundary(this::Boundary{T}, index::Int)::Tuple{bool,T} where {T}
    if this.dim != 1
        error("This function is only available for 1D grids.")
    end

    if index > this.cols
        error("Index is out of bounds.")
    end

    if haskey(this.inner_boundary, (1, index))
        return (false, -1)
    end

    return (true, this.inner_boundary[(1, index)])
end

# Get inner constant boundary condition in 2D case.
#
# Arguments:
#   row: Row index of the inner constant boundary condition
#   col: Column index of the inner constant boundary condition
# Returns: Pair of boolean (whether constant boundary was set or not) and
#          value of the inner constant boundary condition
@export function getInnerBoundary(this::Boundary{T}, row::Int, col::Int)::Tuple{bool,T} where {T}
    if this.dim != 2
        error("This function is only available for 2D grids.")
    end

    if row > this.rows || col > this.cols
        error("Index is out of bounds.")
    end

    if haskey(this.inner_boundary, (row, col))
        return (false, -1)
    end

    return (true, this.inner_boundary[(row, col)])
end

# Get inner constant boundary conditions of a row as a vector. Can be used for 1D grids (row == 0) or 2D grids.
#
# Arguments:
#   row: Index of the row for which the inner boundary conditions are to be returned.
# Returns: Vector of pairs of boolean (whether constant boundary was set or not) and value of the
#          inner constant boundary condition
@export function getInnerBoundaryRow(
    this::Boundary{T},
    row::Int,
)::AbstractVector{Tuple{Bool,T}} where {T}
    if row > this.rows
        error("Index is out of bounds.")
    end

    if length(this.inner_boundary) == 0
        return [(false, -1) for _ = 1:this.cols]
    end

    row_values::Vector{Tuple{Bool,T}} = []

    for col = 1:this.cols
        append!(row_values, getInnerBoundary(this, row, col))
    end

    return row_values
end

# Get inner constant boundary conditions of a column as a vector. Can only be used for 2D grids.
#
# Arguments:
#   col: Index of the column for which the inner boundary conditions are to be returned.
# Returns: Vector of pairs of boolean (whether constant boundary was set or not) and value of the
# inner constant boundary condition
@export function getInnerBoundaryCol(
    this::Boundary{T},
    col::Int,
)::AbstractVector{Tuple{Bool,T}} where {T}
    if this.dim != 2
        error("This function is only available for 2D grids.")
    end

    if col > this.cols
        error("Index is out of bounds.")
    end

    if length(this.inner_boundary) == 0
        return [(false, -1) for _ = 1:this.rows]
    end

    col_values::Vector{Tuple{bool,T}} = []

    for row = 1:this.rows
        append!(row_values, getInnerBoundary(this, row, col))
    end

    return col_values
end

# Get inner constant boundary conditions as a map. Can be read by setInnerBoundaries.
#
# Returns: Map of inner constant boundary conditions
@export function getInnerBoundaries(this::Boundary{T})::Dict{Tuple{UInt32,UInt32},T} where {T}
    return this.boundaries
end

# Set inner constant boundary conditions as a map. Can be obtained by getInnerBoundaries.
#
# Arguments:
#   inner_boundary: Map of inner constant boundary conditions
@export function setInnerBoundaries!(
    this::Boundary{T},
    inner_boundary::Dict{Tuple{UInt32,UInt32},T},
) where {T}
    this.inner_boundary = inner_boundary
end
