using FLoops
using InlineExports
using LinearAlgebra
using SparseArrays

# Enum defining the Linear Equation solvers
@enum SOLVER begin
    JULIA_BUILTIN_SOLVER
    THOMAS_ALGORITHM_SOLVER # Thomas Algorithm solver; more efficient for tridiagonal matrices
end
@exportinstances SOLVER

# calculates coefficient for boundary in constant case
function calcBoundaryCoeffConstant(
    alpha_center::T,
    alpha_side::T,
    sx::T,
)::Tuple{T,T} where {T<:Number}
    centerCoeff::T = 1 + sx * (calcAlphaIntercell(alpha_center, alpha_side) + 2 * alpha_center)
    sideCoeff::T = -sx * calcAlphaIntercell(alpha_center, alpha_side)

    return (centerCoeff, sideCoeff)
end

# calculates coefficient for boundary in closed case
function calcBoundaryCoeffClosed(
    alpha_center::T,
    alpha_side::T,
    sx::T,
)::Tuple{T,T} where {T<:Number}
    centerCoeff::T = 1 + sx * calcAlphaIntercell(alpha_center, alpha_side)
    sideCoeff::T = -sx * calcAlphaIntercell(alpha_center, alpha_side)

    return (centerCoeff, sideCoeff)
end

# creates coefficient matrix for next time step from alphas in x-direction
function createCoeffMatrix(
    alpha::AbstractMatrix{T},
    bcLeft::AbstractVector{BoundaryElement{T}},
    bcRight::AbstractVector{BoundaryElement{T}},
    inner_bc::AbstractVector{Tuple{Bool,T}},
    numCols::Int,
    rowIndex::Int,
    sx::T,
)::AbstractMatrix{T} where {T}
    type::BC_TYPE = BC_TYPE_CLOSED
    cm::SparseMatrixCSC{T} = spzeros(T, numCols, numCols)

    # left column
    if inner_bc[1][1]
        cm[1, 1] = 1
    else
        type = getBoundaryElementType(bcLeft[rowIndex])

        if type == BC_TYPE_CONSTANT
            centerCoeffTop, rightCoeffTop =
                calcBoundaryCoeffConstant(alpha[1, rowIndex], alpha[2, rowIndex], sx)

            cm[1, 1] = centerCoeffTop
            cm[1, 2] = rightCoeffTop
        elseif type == BC_TYPE_CLOSED
            centerCoeffTop, rightCoeffTop =
                calcBoundaryCoeffClosed(alpha[1, rowIndex], alpha[2, rowIndex], sx)

            cm[1, 1] = centerCoeffTop
            cm[1, 2] = rightCoeffTop
        else
            error("Undefined Boundary Condition Type somewhere on Left or Top!")
        end
    end

    # inner columns
    n::Int = numCols - 1

    for i = 2:n
        if inner_bc[i][1]
            cm[i, i] = 1
            continue
        end

        cm[i, i-1] = -sx * calcAlphaIntercell(alpha[i-1, rowIndex], alpha[i, rowIndex])
        cm[i, i] =
            1 +
            sx * (
                calcAlphaIntercell(alpha[i, rowIndex], alpha[i+1, rowIndex]) +
                calcAlphaIntercell(alpha[i-1, rowIndex], alpha[i, rowIndex])
            )
        cm[i, i+1] = -sx * calcAlphaIntercell(alpha[i, rowIndex], alpha[i+1, rowIndex])
    end

    # right column
    if inner_bc[numCols][1]
        cm[numCols, numCols] = 1
    else
        type = getBoundaryElementType(bcRight[rowIndex])

        if type == BC_TYPE_CONSTANT
            centerCoeffBottom, leftCoeffBottom =
                calcBoundaryCoeffConstant(alpha[numCols, rowIndex], alpha[numCols-1, rowIndex], sx)

            cm[numCols, numCols-1] = leftCoeffBottom
            cm[numCols, numCols] = centerCoeffBottom
        elseif type == BC_TYPE_CLOSED
            centerCoeffBottom, leftCoeffBottom =
                calcBoundaryCoeffClosed(alpha[numCols, rowIndex], alpha[numCols-1, rowIndex], sx)

            cm[numCols, numCols-1] = leftCoeffBottom
            cm[numCols, numCols] = centerCoeffBottom
        else
            error("Undefined Boundary Condition Type somewhere on Right or Bottom!")
        end
    end

    return cm
end

# calculates explicit concentration at boundary in closed case
function calcExplicitConcentrationsBoundaryClosed(
    conc_center::T,
    alpha_center::T,
    alpha_neigbor::T,
    sy::T,
)::T where {T}
    return sy * calcAlphaIntercell(alpha_center, alpha_neigbor) * conc_center +
           (1 - sy * (calcAlphaIntercell(alpha_center, alpha_neigbor))) * conc_center
end

# calculates explicity concentration at boundary in constant case
function calcExplicitConcentrationsBoundaryConstant(
    conc_center::T,
    conc_bc::T,
    alpha_center::T,
    alpha_neighbor::T,
    sy::T,
)::T where {T}
    inter_cell::T = calcAlphaIntercell(alpha_center, alpha_neighbor)
    return sy * inter_cell * conc_center +
           (1 - sy * (inter_cell + alpha_center)) * conc_center +
           sy * alpha_center * conc_bc
end

function createSolutionVector(
    concentrations::AbstractMatrix{T},
    alphaX::AbstractMatrix{T},
    alphaY::AbstractMatrix{T},
    bcLeft::AbstractVector{BoundaryElement{T}},
    bcRight::AbstractVector{BoundaryElement{T}},
    bcTop::AbstractVector{BoundaryElement{T}},
    bcBottom::AbstractVector{BoundaryElement{T}},
    inner_bc::AbstractVector{Tuple{Bool,T}},
    length::Int,
    rowIndex::Int,
    sx::T,
    sy::T,
)::AbstractVector{T} where {T}
    sv = Vector{T}(undef, length)
    numRows::Int = size(concentrations, 2)

    type::BC_TYPE = BC_TYPE_CLOSED

    # inner rows
    if rowIndex > 1 && rowIndex < numRows
        for i = 1:length
            if inner_bc[i][1]
                sv[i] = inner_bc[i][2]
                continue
            end

            sv[i] =
                sy *
                calcAlphaIntercell(alphaY[i, rowIndex], alphaY[i, rowIndex+1]) *
                concentrations[i, rowIndex+1] +
                (
                    1 -
                    sy * (
                        calcAlphaIntercell(alphaY[i, rowIndex], alphaY[i, rowIndex+1]) +
                        calcAlphaIntercell(alphaY[i, rowIndex-1], alphaY[i, rowIndex])
                    )
                ) * concentrations[i, rowIndex] +
                sy *
                calcAlphaIntercell(alphaY[i, rowIndex-1], alphaY[i, rowIndex]) *
                concentrations[i, rowIndex-1]
        end
    elseif rowIndex == 1 # first row
        for i = 1:length
            if inner_bc[i][1]
                sv[i] = inner_bc[i][2]
                continue
            end

            type = getBoundaryElementType(bcTop[i])

            if type == BC_TYPE_CONSTANT
                sv[i] = calcExplicitConcentrationsBoundaryConstant(
                    concentrations[i, rowIndex],
                    getBoundaryElementValue(bcTop[i]),
                    alphaY[i, rowIndex],
                    alphaY[i, rowIndex+1],
                    sy,
                )
            elseif type == BC_TYPE_CLOSED
                sv[i] = calcExplicitConcentrationsBoundaryClosed(
                    concentrations[i, rowIndex],
                    alphaY[i, rowIndex],
                    alphaY[i, rowIndex+1],
                    sy,
                )
            else
                error("Undefined Boundary Condition Type somewhere on Left or Top!")
            end
        end
    elseif rowIndex == numRows # last row
        for i = 1:length
            if inner_bc[i][1]
                sv[i] = inner_bc[i][2]
                continue
            end

            type = getBoundaryElementType(bcBottom[i])

            if type == BC_TYPE_CONSTANT
                sv[i] = calcExplicitConcentrationsBoundaryConstant(
                    concentrations[i, rowIndex],
                    getBoundaryElementValue(bcBottom[i]),
                    alphaY[i, rowIndex],
                    alphaY[i, rowIndex-1],
                    sy,
                )
            elseif type == BC_TYPE_CLOSED
                sv[i] = calcExplicitConcentrationsBoundaryClosed(
                    concentrations[i, rowIndex],
                    alphaY[i, rowIndex],
                    alphaY[i, rowIndex-1],
                    sy,
                )
            else
                error("Undefined Boundary Condition Type somewhere on Right or Bottom!")
            end
        end
    end

    # first column -> additional fixed concentration change from perpendicular dimension in constant bc case
    if getBoundaryElementType(bcLeft[rowIndex]) == BC_TYPE_CONSTANT && !inner_bc[1][1]
        sv[1] += 2 * sx * alphaX[1, rowIndex] * getBoundaryElementValue(bcLeft[rowIndex])
    end

    # last column -> additional fixed concentration change from perpendicular dimension in constant bc case
    if getBoundaryElementType(bcRight[rowIndex]) == BC_TYPE_CONSTANT && !inner_bc[length][1]
        sv[length] += 2 * sx * alphaX[length, rowIndex] * getBoundaryElementValue(bcRight[rowIndex])
    end

    return sv
end

# solver for linear equation system; A corresponds to coefficient matrix,  b to the solution vector
# use of the builtin Julia solver
function JuliaBuiltinAlgorithm(
    A::AbstractMatrix{T},
    b::AbstractVector{T},
)::AbstractVector{T} where {T}
    return factorize(A) \ b
end

# solver for linear equation system; A corresponds to coefficient matrix, b to the solution vector
# implementation of Thomas Algorithm
function ThomasAlgorithm(
    A::AbstractMatrix{T},
    x_vec::AbstractVector{T},
)::AbstractVector{T} where {T}
    n::Int = size(x_vec, 1)

    a_diag = Vector{T}(undef, n)
    b_diag = Vector{T}(undef, n)
    c_diag = Vector{T}(undef, n)

    # Fill diagonals vectors
    b_diag[1] = A[1, 1]
    c_diag[1] = A[1, 2]

    for i = 2:n-1
        a_diag[i] = A[i, i-1]
        b_diag[i] = A[i, i]
        c_diag[i] = A[i, i+1]
    end

    a_diag[n] = A[n, n-1]
    b_diag[n] = A[n, n]

    # start solving - c_diag and x_vec are overwritten
    n -= 1

    c_diag[1] = c_diag[1] / b_diag[1]
    x_vec[1] = x_vec[1] / b_diag[1]

    for i = 2:n
        c_diag[i] /= b_diag[i] - a_diag[i] * c_diag[i-1]
        x_vec[i] = (x_vec[i] - a_diag[i] * x_vec[i-1]) / (b_diag[i] - a_diag[i] * c_diag[i-1])
    end

    x_vec[n+1] = (x_vec[n+1] - a_diag[n+1] * x_vec[n]) / (b_diag[n+1] - a_diag[n+1] * c_diag[n])

    for i = 0:n-1
        x_vec[n-i] -= c_diag[n-i] * x_vec[n-i+1]
    end

    return x_vec
end

# BTCS solution for 1D grid
function BTCS_1D!(grid::Grid{T}, bc::Boundary{T}, timestep::T, solver::SOLVER) where {T}
    length::Int = grid_getLength(grid)
    sx::T = timestep / (getGridDelta(grid) * getGridDelta(grid))

    concentrations_t1 = Vector{T}(undef, length)
    v = Vector{T}(undef, length)

    alpha = getGridAlpha(grid)

    bcLeft = getBoundarySide(bc, BC_SIDE_LEFT)
    bcRight = getBoundarySide(bc, BC_SIDE_RIGHT)

    inner_bc = getInnerBoundaryRow(bc, 1)

    concentrations::AbstractMatrix{T} = getGridConcentrations(grid)
    rowIndex::Int = 1

    solverFunc::Function

    if solver == JULIA_BUILTIN_SOLVER
        solverFunc = JuliaBuiltinAlgorithm
    elseif solver == THOMAS_ALGORITHM_SOLVER
        solverFunc = ThomasAlgorithm
    else
        error("Invalid solver type!")
    end

    A::AbstractMatrix{T} = createCoeffMatrix(alpha, bcLeft, bcRight, inner_bc, length, rowIndex, sx) # this is exactly same as in 2D

    for i = 1:length
        b[i] = concentrations[i, 0]
    end

    if getBoundaryElementType(bc, BC_SIDE_LEFT, 1) == BC_TYPE_CONSTANT && !inner_bc[1][1]
        b[1] += 2 * sx * alpha[1, 1] * getBoundaryElementValue(bcLeft[1])
    end

    if getBoundaryElementType(bc, BC_SIDE_RIGHT, 1) == BC_TYPE_CONSTANT && !inner_bc[length][1]
        b[length] += 2 * sx * alpha[length, 1] * getBoundaryElementValue(bcRight[1])
    end

    concentrations = solverFunc(A, b)

    # for j = 1:length
    #     concentrations[j, 1] = concentrations_t1[j]
    # end

    # grid_setConcentrations!(grid, concentrations)
end

# BTCS solution for 2D grid
function BTCS_2D!(grid::Grid{T}, bc::Boundary{T}, timestep::T, solver::SOLVER) where {T}
    rowMax::Int = getGridRow(grid)
    colMax::Int = getGridCol(grid)

    sx::T = timestep / (2 * getGridDeltaCol(grid) * getGridDeltaCol(grid))
    sy::T = timestep / (2 * getGridDeltaRow(grid) * getGridDeltaRow(grid))

    concentrations_t1 = Matrix{T}(undef, (colMax, rowMax))

    alphaX::AbstractMatrix{T} = getGridAlphaX(grid)
    alphaY::AbstractMatrix{T} = getGridAlphaY(grid)

    bcLeft = getBoundarySide(bc, BC_SIDE_LEFT)
    bcRight = getBoundarySide(bc, BC_SIDE_RIGHT)
    bcTop = getBoundarySide(bc, BC_SIDE_TOP)
    bcBottom = getBoundarySide(bc, BC_SIDE_BOTTOM)

    concentrations::AbstractMatrix{T} = getGridConcentrations(grid)

    solverFunc::Function = function (A, b) end

    if solver == JULIA_BUILTIN_SOLVER
        solverFunc = JuliaBuiltinAlgorithm
    elseif solver == THOMAS_ALGORITHM_SOLVER
        solverFunc = ThomasAlgorithm
    else
        error("Invalid solver type!")
    end

    let alphaX = alphaX,
        alphaY = alphaY,
        concentrations_t1 = concentrations_t1,
        solverFunc = solverFunc

        @floop for i = 1:rowMax
            inner_bc = getInnerBoundaryRow(bc, i)

            A = createCoeffMatrix(alphaX, bcLeft, bcRight, inner_bc, colMax, i, sx)
            b = createSolutionVector(
                concentrations,
                alphaX,
                alphaY,
                bcLeft,
                bcRight,
                bcTop,
                bcBottom,
                inner_bc,
                colMax,
                i,
                sx,
                sy,
            )

            concentrations_t1[:, i] = solverFunc(A, b)
        end
    end

    concentrations_t1 = transpose(concentrations_t1)
    alphaX = transpose(alphaX)
    alphaY = transpose(alphaY)

    let alphaX = alphaX,
        alphaY = alphaY,
        concentrations_t1 = concentrations_t1,
        solverFunc = solverFunc

        @floop for i = 1:colMax
            inner_bc = getInnerBoundaryCol(bc, i)

            # swap alphas, boundary conditions and sx/sy for column-wise calculation
            A = createCoeffMatrix(alphaY, bcTop, bcBottom, inner_bc, rowMax, i, sy)
            b = createSolutionVector(
                concentrations_t1,
                alphaY,
                alphaX,
                bcTop,
                bcBottom,
                bcLeft,
                bcRight,
                inner_bc,
                rowMax,
                i,
                sy,
                sx,
            )

            concentrations[i, :] = solverFunc(A, b)
        end
    end
end

# differentiate between 1D and 2D grid
@export function BTCS!(grid::Grid{T}, bc::Boundary{T}, timestep::T, solver::SOLVER) where {T}
    if getGridDim(grid) == 1
        BTCS_1D!(grid, bc, timestep, solver)
    elseif getGridDim(grid) == 2
        BTCS_2D!(grid, bc, timestep, solver)
    else
        error("Error: Only 1- and 2-dimensional grids are defined!")
    end
end

# entry point for Julia solver
@export function BTCS_Julia!(grid::Grid{T}, bc::Boundary{T}, timestep::T) where {T}
    BTCS!(grid, bc, timestep, JULIA_BUILTIN_SOLVER)
end

@export function BTCS_Thomas!(grid::Grid{T}, bc::Boundary{T}, timestep::T) where {T}
    BTCS!(grid, bc, timestep, THOMAS_ALGORITHM_SOLVER)
end
