using InlineExports
using Printf

# Enum defining the implemented solution approaches.
@enum APPROACH begin
    FTCS_APPROACH           # Forward Time-Centered Space
    BTCS_APPROACH           # Backward Time-Centered Space
    CRANK_NICOLSON_APPROACH # Crank-Nicolson method
end
@exportinstances APPROACH

# Enum holding different options for .csv output.
@enum CSV_OUTPUT begin
    CSV_OUTPUT_OFF      # do not produce csv output
    CSV_OUTPUT_ON       # produce csv output with last concentration matrix
    CSV_OUTPUT_VERBOSE  # produce csv output with all concentration matrices
    CSV_OUTPUT_XTREME   # csv output like VERBOSE but additional boundary conditions at beginning
end
@exportinstances CSV_OUTPUT

# Enum holding different options for console output.
@enum CONSOLE_OUTPUT begin
    CONSOLE_OUTPUT_OFF     # do not print any output to console
    CONSOLE_OUTPUT_ON      # print before and after concentrations to console
    CONSOLE_OUTPUT_VERBOSE # print all concentration matrices to console
end
@exportinstances CONSOLE_OUTPUT

# Enum holding different options for time measurement.
@enum TIME_MEASURE begin
    TIME_MEASURE_OFF # do not print any time measures
    TIME_MEASURE_ON  # print time measure after last iteration
end
@exportinstances TIME_MEASURE

# The class forms the interface for performing the diffusion simulations and contains all the
# methods for controlling the desired parameters, such as time step, number of simulations, etc.
#
# T is the type of the internal data structures for grid, boundary condition and timestep
@export mutable struct Simulation{T<:Real}
    approach::APPROACH
    solver::SOLVER

    timestep::T
    iterations::Int
    innerIterations::Int
    csv_output::CSV_OUTPUT
    console_output::CONSOLE_OUTPUT
    time_measure::TIME_MEASURE

    grid::Grid{T}
    bc::Boundary{T}

    approach_names::Dict{APPROACH,String}

    # Set up a simulation environment. The timestep and number of iterations must be set. For the
    # BTCS approach, the Thomas algorithm is used as the default linear equation solver as this is
    # faster for tridiagonal matrices. CSV output, console output and time measure are off by default.
    #
    # Arguments:
    #   grid: Valid grid object
    #   bc: Valid boundary condition object
    #   approach: Set the SLE scheme to be used
    #   solver: Set the solver to be used
    function Simulation{T}(
        grid::Grid{T},
        bc::Boundary{T},
        ;
        approach::APPROACH = BTCS_APPROACH,
        solver::SOLVER = THOMAS_ALGORITHM_SOLVER,
    ) where {T<:Real}
        new(
            approach,
            solver,
            -1,
            -1,
            1,
            CSV_OUTPUT_OFF,
            CONSOLE_OUTPUT_OFF,
            TIME_MEASURE_OFF,
            grid,
            bc,
            Dict(
                FTCS_APPROACH => "FTCS",
                BTCS_APPROACH => "BTCS",
                CRANK_NICOLSON_APPROACH => "CRNI",
            ),
        )
    end
end

# Set the option to output the results to a CSV file. Off by default.
#
# Arguments:
#   csv_output: Valid output option.
@export function setSimulationOutputCSV!(this::Simulation{T}, csv_output::CSV_OUTPUT) where {T}
    this.csv_output = csv_output
end

# Set the options for outputting information to the console. Off by default.
#
# Arguments:
#   console_output: Valid output option.
@export function setSimulationOutputConsole!(
    this::Simulation{T},
    console_output::CONSOLE_OUTPUT,
) where {T}
    this.console_output = console_output
end

# Set the Time Measure option. Off by default.
@export function setSimulationTimeMeasure!(
    this::Simulation{T},
    time_measure::TIME_MEASURE,
) where {T}
    this.time_measure = time_measure
end

# Setting the time step for each iteration step. Time step must be greater than zero.
# Setting the timestep is required.
#
# Arguments:
#   timestep: Valid timestep greater than zero.
@export function setSimulationTimestep!(this::Simulation{T}, timestep::T) where {T}
    if timestep <= 0
        error("Timestep has to be greater than zero.")
    end

    if this.approach == FTCS_APPROACH || this.approach == CRANK_NICOLSON_APPROACH
        cfl::T
        maxAlpha::T

        if getGridDim(this.grid) == 1
            deltaSquare::T = getGridDelta(grid)
            maxAlpha = findmax(getGridAlpha(grid))[1]

            # Courant-Friedrichs-Lewy condition
            cfl = deltaSquare / (4 * maxAlpha)
        elseif grid_getDim(grid) == 2
            deltaColSquare::T = getGridDeltaCol(grid) * getGridDeltaCol(grid)

            # will be 0 if 1D, else ...
            deltaRowSquare::T = getGridDeltaRow(grid) * getGridDeltaRow(grid)
            minDeltaSquare::T = min(deltaColSquare, deltaRowSquare)

            maxAlpha = max(findmax(getGridAlphaX(grid))[1], findmax(getGridAlphaY(grid))[1])
            cfl = minDeltaSquare / (4 * maxAlpha)
        end

        dim::String = @sprintf "%dD" getGridDim(grid)

        apporachPrefix::String = this.approach_names[this.approach]
        println(stdout, apporachPrefix, "_", dim, " :: CFL condition: ", cfl)
        println(stdout, approachPrefix, "_", dim, " :: required dt=", timestep)

        if timestep > cfl
            this.innerIterations = Int(ceil(timestep / cfl))
            this.timestep = timestep / this.innerIterations

            println(
                stderr,
                "Warning :: Timestep was adjusted, because of stability conditions. Time duration was approximately preserved by adjusting internal number of iterations.",
            )
            println(
                stdout,
                apporachPrefix,
                "_",
                dim,
                " :: Required ",
                this.innerIterations,
                " inner iterations with dt=",
                this.timestep,
            )
        else
            this.timestep = timestep
            println(
                stdout,
                apporachPrefix,
                "_",
                dim,
                " :: No inner iterations required, dt=",
                timestep,
            )
        end
    else
        this.timestep = timestep
    end
end

# Currently set time step is returned.
@export function getSimulationTimestep(this::Simulation{T})::T where {T}
    return this.timestep
end

# Set the desired iterations to be calculated. A value greater than zero must be specified here.
# Setting iterations is required.
#
# Arguments:
#   iterations: Number of iterations to be simulated.
@export function setSimulationIterations!(this::Simulation{T}, iterations::Int) where {T}
    if iterations <= 0
        error("Number of iterations must be greater than zero.")
    end

    this.iterations = iterations
end

# Return the currently set iterations to be calculated.
@export function getSimulationIterations(this::Simulation{T})::Int where {T}
    return this.iterations
end

# Outputs the current concentrations of the grid on the console.
function printConcentrationsConsole(this::Simulation{T}) where {T}
    println(stdout, getGridConcentrations(this.grid))
    println(stdout, "")
end

# Creates a CSV file with a name containing the current simulation parameters.
# If the data name already exists, an additional counter is appended to the name.
# The name of the file is built up as follows:
# <approach> + <number rows> + <number columns> + <number of iterations>+<counter>.csv
#
# Returns: Filename with configured simulation parameters.
function createCSVFile(this::Simulation{T})::String where {T}
    appendIdent::Int = 0
    appendIdentString::String = ""

    approachString::String = this.approach_names[this.approach]
    row::String = string(getGridRow(this.grid))
    col::String = string(getGridCol(this.grid))
    numIterations::String = string(this.iterations)

    filename::String = approachString * "_" * row * "_" * col * "_" * numIterations * ".csv"

    while isfile(filename)
        appendIdent += 1
        appendIdentString = string(appendIdent)
        filename =
            approachString *
            "_" *
            row *
            "_" *
            col *
            "_" *
            numIterations *
            "-" *
            appendIdentString *
            ".csv"
    end

    open(filename, "w") do file
        # adds lines at the beginning of verbose output csv that represent the boundary conditions
        # and their values -1 in case of closed boundary
        if this.csv_output == CSV_OUTPUT_XTREME
            sides::Vector{AbstractVector{T}} = [
                getBoundarySideValues(this.bc, BC_SIDE_LEFT),
                getBoundarySideValues(this.bc, BC_SIDE_RIGHT),
                getBoundarySideValues(this.bc, BC_SIDE_TOP),
                getBoundarySideValues(this.bc, BC_SIDE_BOTTOM),
            ]

            for side in sides
                for i in axes(side, 1)
                    @printf file "%f" side[i]

                    if i < size(side, 1)
                        print(file, " ")
                    end
                end

                print(file, "\n")
            end

            print(file, "\n")
            print(file, "\n")
        end
    end

    return filename
end

# Writes the currently calculated concentration values of the grid into the CSV file with the passed filename.
#
# Arguments:
#   filename: Name of the file to which the concentration values are to be written.
function printConcentrationsCSV(this::Simulation{T}, filename::String) where {T}
    open(filename, "a") do file
        concentrations::AbstractMatrix{T} = getGridConcentrations(this.grid)

        for j = axes(concentrations, 2)
            for i = axes(concentrations, 1)
                @printf file "%E" concentrations[i, j]

                if i < size(concentrations, 1)
                    print(file, " ")
                end
            end

            print(file, "\n")
        end

        print(file, "\n")
        print(file, "\n")
    end
end

# Method starts the simulation process with the previously set parameters.
@export function runSimulation!(this::Simulation{T}) where {T}
    if this.timestep == -1
        error("Timestep is not set!")
    end

    if this.iterations == -1
        error("Number of iterations are not set!")
    end

    filename::String = ""

    if this.console_output != CONSOLE_OUTPUT_OFF
        printConcentrationsConsole(this)
    end

    if this.csv_output != CSV_OUTPUT_OFF
        filename = createCSVFile(this)
    end

    milliseconds::Float64 =
        1000 * @elapsed begin
            if this.approach == FTCS_APPROACH
                for i = 1:(this.iterations*this.innerIterations)
                    if this.console_output == CONSOLE_OUTPUT_VERBOSE && i > 1
                        printConcentrationsConsole(this)
                    end

                    if this.csv_output == CSV_OUTPUT_VERBOSE || this.csv_output == CSV_OUTPUT_XTREME
                        printConcentrationsCSV(this, filename)
                    end

                    # FTCS!(this.grid, this.bc, this.timestep)
                end
            elseif this.approach == BTCS_APPROACH
                for i = 1:this.iterations
                    if this.console_output == CONSOLE_OUTPUT_VERBOSE && i > 1
                        printConcentrationsConsole(this)
                    end

                    if this.csv_output == CSV_OUTPUT_VERBOSE || this.csv_output == CSV_OUTPUT_XTREME
                        printConcentrationsCSV(this, filename)
                    end

                    BTCS!(this.grid, this.bc, this.timestep, this.solver)
                end
            elseif this.approach == CRANK_NICOLSON_APPROACH
                beta::T = 0.5

                # TODO this implementation is very inefficient!
                # a separate implementation that sets up a specific tridiagonal matrix
                # for Crank-Nicolson would be better

                for i = 1:(this.iterations*this.innerIterations)
                    if this.console_output == CONSOLE_OUTPUT_VERBOSE && i > 1
                        printConcentrationsConsole(this)
                    end

                    if this.csv_output == CSV_OUTPUT_VERBOSE || this.csv_output == CSV_OUTPUT_XTREME
                        printConcentrationsCSV(this, filename)
                    end

                    concentrations::AbstractMatrix{T} = getGridConcentrations(this.grid)
                    # FTCS!(this.grid, this.bc, this.timestep)
                    concentrationsFTCS::AbstractMatrix{T} = getGridConcentrations(this.grid)
                    setGridConcentrations!(this.grid, concentrations)
                    BTCS!(this.grid, this.bc, this.timestep, THOMAS_ALGORITHM_SOLVER)
                    concentrationsResult::AbstractMatrix{T} =
                        beta * concentrationsFTCS + (1 - beta) * getGridConcentrations(this.grid)
                    setGridConcentrations!(this.grid, concentrationsResult)
                end
            end

            if this.console_output != CONSOLE_OUTPUT_OFF
                printConcentrationsConsole(this)
            end

            if this.csv_output != CSV_OUTPUT_OFF
                printConcentrationsCSV(this, filename)
            end

            if this.time_measure != TIME_MEASURE_OFF
                approachString::String = this.approach_names[this.approach]
                dimString::String = @sprintf "%dD" getGridDim(this.grid)
                println(
                    stdout,
                    approachString,
                    dimString,
                    ":: run() finished in ",
                    Int(milliseconds),
                    "ms",
                )
            end
        end
end
